(* Proof of fixpoint combinator example for
   Realizability and Compositional Compiler Correctness 
      for a Polymorphic Language
   N Benton & C-K Hur
   April 2010
*)

Require Import Arith List Program FunctionalExtensionality Heq.
Require Import util.
Require Import ssreflect ssrbool Heq.
Require Import cesd sysf sysfopsem logrel compcorrect.

Set Implicit Arguments.

(******
   Fixpoint Combinator
 ******)

Module EgFixCombinator (ObsLeMod: ObsLeSig).

Module ObsLeModInstance := ObsLeMod.
Module LogRelMod := LogicalRelationFunctor (ObsLeModInstance).
Export LogRelMod.
Module CompCorrect := CompilerCorrectFunctor (ObsLeModInstance).
Export CompCorrect.

(*
   RecSrc 
*)

Program Definition RecSrcOpen 
  : Val 2 nil (((#TV 1 --> #TV 0) --> #TV 1 --> #TV 0) --> #TV 1 --> #TV 0)
  :=
  LAM (VAL                          (* lam F. *)
    (REC                            (*   rec f a. *)
      (LET (APP (#V 2) (#V 1))    (*     let Ff := APP F f in *)
        (APP (#V 0) (#V 1))))).   (*       APP Ff a *)

Program Definition RecSrc 
  : CVal (ALL (ALL (((#TV 1-->#TV 0) --> #TV 1 --> #TV 0) --> #TV 1 --> #TV 0)))
  :=
  TLAM (TLAM (RecSrcOpen)).         (* LAM. LAM. RecSrcOpen *)


(*
   CESD code for fixpoint combinator
*)

(* lam F. lam x. (lam y. F (lam z. y y z)) (lam y. F (lam z. y y z)) x *)
Definition MFixC : Code :=
  [PushC [
    PushC [
      PushC [Push 2; PushC [Push 1; Push 1; App; Push 0; App; Ret]; App; Ret];
      PushC [Push 2; PushC [Push 1; Push 1; App; Push 0; App; Ret]; App; Ret];
      App;
      Push 0;
      App;
      Ret
    ];
    Ret
  ]].

(*
   Correctness Le
*)

Program Definition RECSrc_F (sub: subT 2 0) 
  (F: CVal (subTy sub ((#TV 1 --> #TV 0) --> #TV 1 --> #TV 0))) 
  : CVal (subTy sub (#TV 1 --> #TV 0)) 
  :=
  REC (LET (APP (UPVal [_;_] F) (#V 1)) (APP (#V 0) (#V 1))).

Lemma MFixC_le_core: forall gi (srel: iSubRel 2) F TF, 
  isr_decr srel -> irelLeVFun srel gi F TF -> 
  forall i v e, i <= gi ->
    irelLeVFun srel
    i
    (VC
      (VC (v :: F :: e) 
        [Push 2; PushC [Push 1; Push 1; App; Push 0; App; Ret]; App; Ret]
        :: v :: F :: e)
      [Push 1; Push 1; App; Push 0; App; Ret])
    (RECSrc_F (sub:= (isr_sub srel)) TF).
Proof.
induction i; intros; first by apply irelLeVFun_arr_zero.
do 4 red; intros. simpl. fold irelLeVFun in *.
destruct k0. apply stepsfor_zero.
eapply stepsfor_step_strict; [ constructor | simpl pred ].
ssteps. change [App; Ret] with ([App]++[Ret]). rewrite (app_nil_end [_; F]).
eapply H0. 
- instantiate (1:= k0). omega.
- fold irelLeVFun. rewrite irelLeVFun_fold. 
  apply irelLeVFun_decr with i; by [ | omega | apply IHi; omega].
- instantiate (1:= k0). fold irelLeVFun. rewrite irelLeVFun_fold. 
  red; intros. destruct H7 as [tvv [[nstp ?] ?]].
  ssteps. rewrite (app_nil_end [_; _]).
  eapply H10.
  - apply H8.
  - rewrite irelLeVFun_fold. 
    eapply irelLeVFun_decr with k; by [ | omega | apply H3].
  - instantiate (1:= k1). fold irelLeVFun. red; intros. 
    destruct H11 as [tvv' [[nstp' ?] ?]]. ssteps.
    eapply H4.
    - eexists. split.
      - eexists. unfold RECSrc_F. eapply e_RApp. unfold_SubVE. eapply e_Let.
        - revert H7; unfold RECSrc_F; simpl_Sub; intro; apply H7.
        - unfold_SubVE; simpl_Sub. apply H11.
      - apply H14.
    - omega. - omega.
  - omega. - omega.
- omega. - omega. 
Qed.

Lemma MFixC_le: #c MFixC <<cc<< ^RecSrc.
Proof.
apply cv_implies_cc_le.
eexists. split; first by unfold cstr; apply ObsLeV_refl.
unfold CtoC, MFixC. red; intros. destruct cesd as [[[c0 e0] s0] d0].
eexists. split; first by eexists; simpl; ksteps.
do 4 red; unfold VtoC; intros. fold irelLeVFun in *. eapply H4.
- unfold RecSrc, RecSrcOpen. eexists. split.
  - unfold_SubVE. fold_fix (subSub shsub te) ltac:(fun X => red in X).
    Hhide_evars. Hsimplc. Hshow_evars. eexists.
    by apply (@e_TApp (ALL (((#TV 1 -->  #TV 0) --> #TV 1 -->  #TV 0) --> 
      #TV 1 -->  #TV 0)) _ t).
  - instantiate (1:= k1). do 4 red; intros. eapply H8.
    - assert (C0 : (nil:Env 0, ((subTy [|t0|] (renTy shren t) -->  t0) --> 
      subTy [|t0|] (renTy shren t) --> t0) --> subTy [|t0|] (renTy shren t) --> 
      t0) = (nil, ((t -->  t0) -->  t -->  t0) -->  t -->  t0))
       by (simpl_sub).
      eexists. split.
      - Hhide_evars. unfold_SubVE. Hsimpl. Helimc -> C. Hrefl (hev0 : SVal PVal).
        (* manual *) Hget C2. split_tuple. hrewritec (proj2 H11). clear H11.
        intros; eapply eq_ind; first by eexists; 
         apply (@e_TApp (((renTy shren t --> #TV 0) --> renTy shren t --> #TV 0)
          --> renTy shren t --> #TV 0) _ t0).
        unfold_SubVE. Hshow_evars. instantiate (1:= << SVal,PVal # C0 >> _).
        Hhide_evars. Hsimpl; Hclear. Hshow_evars. by reflexivity.
      - instantiate (1:= k2). do 4 red; intros. fold irelLeVFun in *.
        simpl. eapply stepsfor_step_strict; first by constructor.
        ssteps. eapply H13.
        - eexists. split.
          - Hhide_evars. Hsimpl. Helimc -> C. Hshow_evars. eexists.
            constructor. unfold_SubVE. by constructor.
          - instantiate (1:= k4). do 3 red; intros. simpl.
            ssteps. change [App;Ret] with ([App]++[Ret]). 
            rewrite (app_nil_end [_;_]). eapply H12.
            - instantiate (1:= k6). by omega.
            - eapply MFixC_le_core.
              - by do 2 (apply isr_cons_decr; auto).
              - by apply H12.
            - by omega.
            - instantiate (1:= k6). red; intros. 
              destruct H21 as [tvv [[nstp ?] ?]].
              ssteps. rewrite (app_nil_end [_;_]). eapply H24.
              - by apply H22.
              - rewrite irelLeVFun_fold. 
                apply irelLeVFun_decr with k5; 
                 by [do 2 (apply isr_cons_decr; auto)|omega|Apply].
              - instantiate (1:= k7). rewrite irelLeVFun_fold. 
                red; intros. destruct H25 as [tvv' [[nstp' ?] ?]].
                ssteps. eapply H18.
                - eexists. split.
                  - eexists. constructor. unfold_SubVE. econstructor. 
                    - revert H21. unfold RECSrc_F. simpl_Sub. intro. apply H21.
                    - unfold_SubVE. simpl_Sub. apply H25.
                  - apply H28.
omega. omega. omega. omega. omega. omega.
omega. omega. omega. omega. omega. omega.
Qed.

(*
   Correctness Ge
*)

Program Definition RECNSrc_F (sub: subT 2 0) i 
  (F: CVal (subTy sub ((#TV 1 --> #TV 0) --> #TV 1 --> #TV 0)))
  : CVal (subTy sub (#TV 1 --> #TV 0)) 
  :=
  RECN i (LET (APP (UPVal [_;_] F) (#V 1)) (APP (#V 0) (#V 1))).

Lemma MFixC_ge_core: forall (srel: SubRel 2) F TF, erelGeVFun srel F TF -> 
  forall i v e, 
    erelGeVFun srel
    (VC
      (VC (v :: F :: e) 
        [Push 2; PushC [Push 1; Push 1; App; Push 0; App; Ret]; App; Ret]
        :: v :: F :: e)
      [Push 1; Push 1; App; Push 0; App; Ret])
    (RECNSrc_F (sub:= sr_sub srel) i TF).
Proof.
induction i; intros; first by apply BotF_bot.
do 4 red; intros. simpl. fold erelGeVFun in *.
tsteps. destruct H1. destruct x; first by inversion H1.
replace_hyp H1 (Ev_to_Treduce (Ev_step_rev H1 (step_App _ _))). fold RECN in H1.
revert H1. unfold_SubVE. intro. 
destruct (Treduce_let_inv H1) as [k' [tv' [? [? ?]]]]. clear H1.
revert H4; simpl_Sub; intro. replace_hyp H4 (Treduce_to_Ev H4).
revert H5. unfold_SubVE. simpl_Sub. intro. replace_hyp H5 (Treduce_to_Ev H5).
change [App; Ret] with ([App]++[Ret]). rewrite (app_nil_end [_; F]). eapply H.
- rewrite erelGeVFun_fold. apply IHi.
- unfold RECNSrc_F, UPVal. rewrite SubVal_SS. eexists. apply H4.
- rewrite erelGeVFun_fold. red; intros.
  tsteps. rewrite (app_nil_end [v0;val]). eapply H1.
  - rewrite erelGeVFun_fold. apply H0.
  - eexists. apply H5.
  - rewrite erelGeVFun_fold. red; intros. tsteps. apply H2; auto.
Qed.

Program Definition RecNSrcOpen i 
  : Val 2 nil (((#TV 1 --> #TV 0) --> #TV 1 --> #TV 0) --> #TV 1 --> #TV 0)
  :=
  (LAM (VAL
    (RECN i
      (LET (APP (#V 2) (#V 1))
        (APP (#V 0) (#V 1)))))).

Program Definition RecNSrc i 
  : CVal (ALL (ALL (((#TV 1-->#TV 0)-->(#TV 1-->#TV 0))-->#TV 1-->#TV 0)))
  :=
  TLAM (TLAM (RecNSrcOpen i)).

Lemma MFixC_ge: #c MFixC >>cc>> ^RecSrc.
Proof.
apply cv_implies_cc_ge.
unfold CtoC; red. exists (fun i => RecNSrc (S i)). split.
- unfold cstr; apply ObsLe_TLAM; unfold cstr. 
  apply ObsLe_TLAM; unfold cstr; apply ObsLe_LAM, ObsLe_VAL.
  eapply ObsLeV_trans; first by apply ObsLe_unwinding.
  intros; eapply ObsLeV_trans. 
  - apply RECN_monotone; first by instantiate (1:= S p); auto with arith. 
    by apply ObsLeC_refl.
  - intros; unfold cstr; apply (ObsLeV_elmt _ (tvs:= fun _ => _)).
    red; intros.
    move : RECN_monotone; unfold cstr; apply; by [omega | apply ObsLeC_refl].
- red; intros. destruct cesd as [[[c0 e0] s0] d0]. 
  eexists. split; first by eexists; simpl; ksteps.
  do 4 red. unfold VtoC. intros. rewrite -> erelGeVFun_fold in *. apply H2.
  do 6 red. unfold VtoC. intros. rewrite -> erelGeVFun_fold in *. apply H4.
  do 6 red. unfold VtoC. intros. rewrite -> erelGeVFun_fold in *. simpl.
  tsteps. apply H7.
  do 6 red. unfold VtoC. intros. rewrite -> erelGeVFun_fold in *. simpl.
  tsteps. destruct H1.
  Hhide_concl. revert H1. Helimc. unfold RecNSrc. unfold_SubVE. intro.
  edestruct (Ev_uniq H1); clear H1; 
   first by apply (@e_TApp (subTy (subL (sr_sub srel)) 
    (ALL (((#TV 1 --> #TV O) --> #TV 1 --> #TV O) --> #TV 1 --> #TV O))) _ t).
  rewrite -> H12 in *. clear H11 H12 tv.
  revert H3. Hsimplc. unfold_SubVE. Hsimpl. Helimc -> C. Hrefl (tv0 : SVal PVal).
  (* manual *) Hget C2. split_tuple. hrewritec (proj2 H1). clear H1. intros.
  destruct H1. edestruct (Ev_uniq H1); clear H1; 
   first by apply (@e_TApp (((renTy shren t --> #TV 0) --> renTy shren t
     --> #TV 0) --> renTy shren t --> #TV 0) _ t0).
  apply (eq_Heq SVal PVal) in H11. 
  Hsimpl_hyps. rewrite -> (Hrw H11) in *. clear H3 C.
  revert H6. Hsimpl. unfold_SubVE. 
  Hrefl (tv1 : SVal PVal). Hrefl (tv2 : SVal PVal). unfold sr_sub; simpl. 
  (* manual *) Hgeteq C1. hrewrite H6 at 4; hrewrite H6 at 5; hrewritec H6 at 6.
  clear H12. intros.
  destruct H1. edestruct (Ev_uniq H1); clear H1; 
   first by constructor; unfold_SubVE; constructor.
  apply (eq_Heq SVal PVal) in H6. Hsimpl_hyps. 
  rewrite -> (Hrw H6) in *. clear x H3 C0.
  revert H9. Hsimpl. Hrefl (tv3 : SVal PVal). unfold sr_sub; simpl.
  (* manual *) Hgeteq C3. hrewritec H9 at 2. intros.
  destruct H1. destruct x; first by inversion H1.
  replace_hyp H1 (Ev_step_rev H1 (step_App _ _)).
  revert H1. unfold_SubVE. intros. 
  destruct (Treduce_let_inv (Ev_to_Treduce H1)) as [k'' [tv'' [? [? ?]]]]. 
  revert H9 H12. unfold_SubVE. 
  rewrite -> !RenVal_def, !SubVal_SS, !consMap_shSub, 
   !SubVal_idSub, !consMap_SubL, !idSub_SS. 
  repeat rewrite -> ?SubRECN, ?subRECN. unfold_SubVE. 
  rewrite -> !RenVal_def, !SubVal_SS.
  Hsimpl. Helimc C4. Hrefl (tv'' : SVal PVal). Helimc C5. intros. Hshow_concl.
  change [App; Ret] with ([App]++[Ret]). rewrite (app_nil_end [_;_]). eapply H5.
  - rewrite erelGeVFun_fold. eapply MFixC_ge_core. by apply H5.
  - eexists. unfold RECNSrc_F, UPVal. rewrite !SubVal_SS. 
    by apply (Treduce_to_Ev H9).
  - rewrite erelGeVFun_fold. red; intros.
    tsteps. rewrite (app_nil_end [_;_]). eapply H13.
    - apply H8.
    - eexists. replace_hyp H12 (Treduce_to_Ev H12).
      revert H12. Hhide_evars. Hunify. intros. Hshow_evars. 
      by apply H12.
    - rewrite erelGeVFun_fold. red; intros. tsteps. auto.
Qed.

(*
   Correctness
*)

Theorem MFixC_eq: #c MFixC ==cc== ^RecSrc.
Proof. split. apply MFixC_le. apply MFixC_ge. Qed.

End EgFixCombinator.
